/*
 * Copyright (c) 2015 by inovat, innovative systeme - verkehr - tunnel - technik, Dipl.-Ing. H. C. Kniss
 *
 * This file is part of de.bsvrz.kex.tls.TlsApplication
 *
 * de.bsvrz.kex.tls.TlsApplication is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  de.bsvrz.kex.tls.TlsApplication is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with de.bsvrz.kex.tls.TlsApplication.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Diese Datei ist Teil von de.bsvrz.kex.tls.TlsApplication.
 *
 *  de.bsvrz.kex.tls.TlsApplication ist Freie Software: Sie knnen es unter den Bedingungen
 *  der GNU General Public License, wie von der Free Software Foundation,
 *  Version 3 der Lizenz oder (nach Ihrer Wahl) jeder spteren
 *  verffentlichten Version, weiterverbreiten und/oder modifizieren.
 *
 *  de.bsvrz.kex.tls.TlsApplication wird in der Hoffnung, dass es ntzlich sein wird, aber
 *  OHNE JEDE GEWHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 *  Gewhrleistung der MARKTFHIGKEIT oder EIGNUNG FR EINEN BESTIMMTEN ZWECK.
 *  Siehe die GNU General Public License fr weitere Details.
 *
 *  Sie sollten eine Kopie der GNU General Public License zusammen mit diesem
 *  Programm erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * inovat, Dipl.-Ing. H. C. Kniss
 * An der Krautwiese 37
 * D-53783 Eitorf
 * +49 (0)2243 8464 193
 * info@inovat.de
 * www.inovat.de
 */



package de.bsvrz.kex.tls;

//~ NICHT JDK IMPORTE =========================================================

import de.bsvrz.dav.daf.main.ClientDavConnection;
import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientDavParameters;
import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.kex.tls.osi2osi3.osi3.TlsNetworkInitializer;
import de.bsvrz.kex.tls.osi2osi3.osi3.TlsNetworkLayer;
import de.bsvrz.kex.tls.osi7.*;
import de.bsvrz.kex.tls.osi7.deausfallueberwachung.DeAusfallUeberwachung;
import de.bsvrz.kex.tls.osi7.testumgebung.TestSimulationEmpfangeneTelegramme;
import de.bsvrz.pat.onlprot.standardProtocolModule.ProtocolModule;
import de.bsvrz.pat.onlprot.standardProtocolModule.StandardProtocoller;
import de.bsvrz.sys.funclib.commandLineArgs.ArgumentList;
import de.bsvrz.sys.funclib.debug.Debug;

//~ JDK IMPORTE ===============================================================

import java.io.PrintWriter;

//~ KLASSEN ===================================================================

/**
 * TLS-Applikation
 *
 * @author inovat, innovative systeme - verkehr - tunnel - technik
 * @author Dipl.-Ing. Hans Christian Kni (HCK)
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class TlsApplication {
    private static int           _anzahlIntervalleFuerDatenAusfallErkennung;
    private static boolean       _controlLostTelegramsFg4;
    private static boolean       _controlLostTelegramsFg9;
    public static ProtocolModule _protTlsReceived;
    public static ProtocolModule _protTlsSend;
    private static boolean       _sendTimeSync;
    private static int           _standardPeriodenDauerUfdFuerDatenAusfallErkennungInSekunden;
    private static boolean       _timeSyncToDe255;

    //~ FELDER ================================================================

    private final String    _collectDeBlocksClassName;
    private final DataModel _config;

    // Private ----------------------------------------------------------------
    private final ClientDavInterface                        _daf;
    private final DeAusfallUeberwachung                     _deAusfallUeberwachung;
    private final String                                    _deBlocksMultiBlockTransferClassName;
    private final TlsNetworkLayer                           _networkLayer;
    private final String                                    _newRoutingClassName;
    private final String                                    _pidMessageSender;
    private final String                                    _sendingOsi7AdressNullClassName;
    private final boolean                                   _sendOnlyConnectedSM;
    private final boolean                                   _sendOnlyDeOk;
    private final String                                    _sourceDataReferencesClassName;
    private final boolean                                   _testModus;
    private final TestSimulationEmpfangeneTelegramme        _testSimulationEmpfangeneTelegramme;
    private final de.bsvrz.kex.tls.osi7.TlsApplicationLayer _tlsApplicationLayer;

    //~ KONSTRUKTOREN  (und vom Konstruktor verwendete Methoden) ==============

    public TlsApplication(String[] commandArguments) throws Exception {
        ArgumentList argumentList = new ArgumentList(commandArguments);

        Debug.init("KEx-TLS", argumentList);

        Debug debug = Debug.getLogger();

        Thread.setDefaultUncaughtExceptionHandler(new UncaughtExceptionHandler());

        ClientDavParameters parameters = new ClientDavParameters(commandArguments);

        parameters.setApplicationName("KEx-TLS");

        String localDevicePid = null;

        if (argumentList.hasArgument("-gert")) {
            localDevicePid = argumentList.fetchArgument("-gert").asNonEmptyString();
        } else if (argumentList.hasArgument("-geraet")) {
            localDevicePid = argumentList.fetchArgument("-geraet").asNonEmptyString();
        } else if (argumentList.hasArgument("-device")) {
            localDevicePid = argumentList.fetchArgument("-device").asNonEmptyString();
        }

        _sourceDataReferencesClassName = argumentList.fetchArgument("-QuellDatenReferenzen=de.bsvrz.kex.tls.osi7.conversion.SourceDataReferencesDefault").asNonEmptyString();
        _collectDeBlocksClassName      = argumentList.fetchArgument("-DeBloeckeFuerEinzelTelegramme=de.bsvrz.kex.tls.osi7.conversion.CollectDeBlocksDefault").asNonEmptyString();

        // Initialisierung der Klasse CollectDeBlock, damit diese dann in den DeConversionModulen verwendet werden kann.
        CollectDeBlocks collectDeBlocks = CollectDeBlocks.getInstance();

        collectDeBlocks.init(_collectDeBlocksClassName);
        _sendingOsi7AdressNullClassName = argumentList.fetchArgument("-DeBloeckeMitOsi7AdresseNull=de.bsvrz.kex.tls.osi7.conversion.SendingOsi7AdressNullDefault").asNonEmptyString();

        // Initialisierung der Klasse SendingOsi7AdressNull, damit diese dann in den BuildOsi7Telegram verwendet werden kann.
        SendingOsi7AdressNull sendingOsi7AdressNull = SendingOsi7AdressNull.getInstance();

        sendingOsi7AdressNull.init(_sendingOsi7AdressNullClassName);
        _newRoutingClassName = argumentList.fetchArgument("-newRouting=de.bsvrz.kex.tls.osi7.conversion.NewRoutingDefault").asNonEmptyString();

        // Initialisierung der Klasse NewRouting, damit diese dann in den BuildOsi7Telegram verwendet werden kann.
        NewRouting newRouting = NewRouting.getInstance();

        newRouting.init(_newRoutingClassName);

        // Gieseler, BitCtrl
        // Initialisierung der Klasse DeBlocksMultiBlockTransfer, damit diese dann in den BuildOsi7Telegram verwendet werden kann.
        _deBlocksMultiBlockTransferClassName = argumentList.fetchArgument("-DeBloeckeMultiBlockTransfer=de.bsvrz.kex.tls.osi7.conversion.DeBlocksMultiBlockTransferDefault").asNonEmptyString();

        DeBlocksMultiBlockTransfer deBlocksMultiBlockTransfer = DeBlocksMultiBlockTransfer.getInstance();

        deBlocksMultiBlockTransfer.init(_deBlocksMultiBlockTransferClassName);

        // Ende Gieseler, BitCtrl
        _sendOnlyDeOk                                                = argumentList.fetchArgument("-nurSendenAnDeOk=nein").booleanValue();
        _sendOnlyConnectedSM                                         = argumentList.fetchArgument("-nurSendenAnVerbundeneSM=nein").booleanValue();
        _controlLostTelegramsFg4                                     = argumentList.fetchArgument("-FolgenummernVerwaltungFg4=nein").booleanValue();
        _controlLostTelegramsFg9                                     = argumentList.fetchArgument("-FolgenummernVerwaltungFg9=nein").booleanValue();
        _testModus                                                   = argumentList.fetchArgument("-TestModus=nein").booleanValue();
        _timeSyncToDe255                                             = argumentList.fetchArgument("-ZeitsyncronisationAnDe255=nein").booleanValue();
        _sendTimeSync                                                = argumentList.fetchArgument("-ZeitsyncronisationVerschicken=ja").booleanValue();
        _anzahlIntervalleFuerDatenAusfallErkennung                   = argumentList.fetchArgument("-anzahlIntervalleFuerDatenAusfallErkennung=0").intValueBetween(0, 999);
        _standardPeriodenDauerUfdFuerDatenAusfallErkennungInSekunden = argumentList.fetchArgument("-standardPeriodenDauerUfdFuerDatenAusfallErkennungInSekunden=60").intValueBetween(0, 43200);

        // Protokollierer fr empfange und versandte Daten initialisieren
        _protTlsReceived = new StandardProtocoller("############################################################################## TLS >>>>>>> DaV");

        PrintWriter ptr = new PrintWriter(System.out, true);

        _protTlsReceived.initProtocol(argumentList, ptr, commandArguments);
        _protTlsSend = new StandardProtocoller("############################################################################## TLS <<<<<<<<< DaV");

        PrintWriter pts = new PrintWriter(System.err, true);

        _protTlsSend.initProtocol(argumentList, pts, commandArguments);
        argumentList.ensureAllArgumentsUsed();
        _daf = new ClientDavConnection(parameters);
        _daf.connect();
        _daf.login();
        _config = _daf.getDataModel();

        // Versenden von Betriebsmeldungen initialisieren
        _pidMessageSender = argumentList.fetchArgument("-pidBetriebsMeldungsVerwaltung=").asString();

        MessageSender messageSender = MessageSender.getInstance();

        messageSender.init(_daf, "KExTlsOsi7", _pidMessageSender);

        // ...
        ConfigurationObject localDevice = (ConfigurationObject) _config.getObject(localDevicePid);

        if (localDevice == null) {
            throw new IllegalArgumentException("Angegebenes Gert ist nicht bekannt: " + localDevicePid);
        }

        if (!localDevice.isOfType("typ.gert")) {
            throw new IllegalArgumentException("Angegebenes Objekt ist kein Gert: " + localDevicePid);
        }

        _networkLayer = new TlsNetworkLayer();
        new TlsNetworkInitializer(_daf, localDevice, _networkLayer);
        _tlsApplicationLayer                = new TlsApplicationLayer(_daf, localDevice, _networkLayer, _sourceDataReferencesClassName, _sendOnlyDeOk, _sendOnlyConnectedSM);
        _deAusfallUeberwachung              = new DeAusfallUeberwachung();
        _testSimulationEmpfangeneTelegramme = new TestSimulationEmpfangeneTelegramme(_tlsApplicationLayer);
    }

    //~ GET METHODEN ==========================================================

    public static int getAnzahlIntervalleFuerDatenAusfallErkennung() {
        return _anzahlIntervalleFuerDatenAusfallErkennung;
    }

    public static int getStandardPeriodenDauerUfdFuerDatenAusfallErkennungInSekunden() {
        return _standardPeriodenDauerUfdFuerDatenAusfallErkennungInSekunden;
    }

    public static boolean isControlLostTelegramsFg4() {
        return _controlLostTelegramsFg4;
    }

    public static boolean isControlLostTelegramsFg9() {
        return _controlLostTelegramsFg9;
    }

    public static boolean isSendTimeSync() {
        return _sendTimeSync;
    }

    public static boolean isTimeSyncToDe255() {
        return _timeSyncToDe255;
    }

    //~ METHODEN ==============================================================

    public static void main(String[] commandArguments) {
        try {
            TlsApplication app = new TlsApplication(commandArguments);

            app.start();
            System.out.println("Weiter gehts##############################################################################");
        }
        catch (Exception e) {
            Debug debug = Debug.getLogger();

            debug.error("Fehler in der Applikation: " + e);
            e.printStackTrace();
            System.exit(1);
        }
    }

    public void start() {
        _networkLayer.start();
        _tlsApplicationLayer.start();
        _deAusfallUeberwachung.start();

        if (_testModus) {
            _testSimulationEmpfangeneTelegramme.start();
        }
    }

    //~ INNERE KLASSEN ========================================================

    /**
     * Implementierung eines UncaughtExceptionHandlers, der bei nicht abgefangenen Exceptions und Errors entsprechende Ausgaben macht und im Falle eines Errors den
     * Prozess terminiert.
     */
    private static class UncaughtExceptionHandler implements Thread.UncaughtExceptionHandler {

        /**
         * Speicherreserve, die freigegeben wird, wenn ein Error auftritt, damit die Ausgaben nach einem OutOfMemoryError funktionieren
         */
        private volatile byte[] _reserve = new byte[20000];

        //~ METHODEN ==========================================================

        public void uncaughtException(Thread t, Throwable e) {
            if (e instanceof Error) {

                // Speicherreserve freigeben, damit die Ausgaben nach einem OutOfMemoryError funktionieren
                // noinspection AssignmentToNull
                _reserve = null;

                Debug _debug = Debug.getLogger();

                try {
                    System.err.println("Schwerwiegender Laufzeitfehler: Ein Thread hat sich wegen eines Errors beendet, Prozess wird terminiert");
                    System.err.println(t);
                    e.printStackTrace(System.err);
                    _debug.error("Schwerwiegender Laufzeitfehler: " + t + " hat sich wegen eines Errors beendet, Prozess wird terminiert", e);
                }
                catch (Throwable ignored) {

                    // Weitere Probleme whrend der Ausgaben werden ignoriert, damit folgendes exit() auf jeden Fall ausgefhrt wird.
                }

                System.exit(1);
            } else {
                System.err.println("Laufzeitfehler: Ein Thread hat sich wegen einer Exception beendet:");
                System.err.println(t);
                e.printStackTrace(System.err);

                Debug _debug = Debug.getLogger();

                _debug.warning("Laufzeitfehler: " + t + " hat sich wegen einer Exception beendet", e);
            }
        }
    }
}


//~Formatiert mit 'inovat Kodierkonvention' am 30.07.15
