/*
 * Copyright (c) 2015 by inovat, innovative systeme - verkehr - tunnel - technik, Dipl.-Ing. H. C. Kniss
 *
 * This file is part of de.bsvrz.kex.tls.osi7.De
 *
 * de.bsvrz.kex.tls.osi7.De is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  de.bsvrz.kex.tls.osi7.De is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with de.bsvrz.kex.tls.osi7.De.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Diese Datei ist Teil von de.bsvrz.kex.tls.osi7.De.
 *
 *  de.bsvrz.kex.tls.osi7.De ist Freie Software: Sie knnen es unter den Bedingungen
 *  der GNU General Public License, wie von der Free Software Foundation,
 *  Version 3 der Lizenz oder (nach Ihrer Wahl) jeder spteren
 *  verffentlichten Version, weiterverbreiten und/oder modifizieren.
 *
 *  de.bsvrz.kex.tls.osi7.De wird in der Hoffnung, dass es ntzlich sein wird, aber
 *  OHNE JEDE GEWHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 *  Gewhrleistung der MARKTFHIGKEIT oder EIGNUNG FR EINEN BESTIMMTEN ZWECK.
 *  Siehe die GNU General Public License fr weitere Details.
 *
 *  Sie sollten eine Kopie der GNU General Public License zusammen mit diesem
 *  Programm erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * inovat, Dipl.-Ing. H. C. Kniss
 * An der Krautwiese 37
 * D-53783 Eitorf
 * +49 (0)2243 8464 193
 * info@inovat.de
 * www.inovat.de
 */

package de.bsvrz.kex.tls.osi7;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.dav.daf.main.config.SystemObjectType;
import de.bsvrz.kex.tls.osi7.conversion.DeConversionModul;
import de.bsvrz.sys.funclib.debug.Debug;

import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.regex.Pattern;

/**
 * Logisches Verwaltungsobjekt fr ein DE.
 *
 * @author inovat, innovative systeme - verkehr - tunnel - technik
 * @author Dipl.-Ing. Hans Christian Kni (HCK)
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class De {
	// ------------------------------------------------------------------------
	// VARIABLENDEKLARATIONEN
	// ------------------------------------------------------------------------

	// Sonstige ---------------------------------------------------------------
	/** DebugLogger fr Debug-Ausgaben */
	private static final Debug debug = Debug.getLogger();
	// Private ----------------------------------------------------------------
	/** Klassenobjekt des UmsetzungsModuls fr die Konvertierung der Telegramme in Attributgruppen (und zurck) */
	private Class _conversionClass;

	/** DaV Interface */
	private final ClientDavInterface _daf;

	/**
	 * Objektreferenz auf das Objekt vom "typ.deXyz" aus dem Teilmodell TLS oder einen beliebigen Typ bei TLS hnlicher
	 * Kommunikation.
	 */
	private final ConfigurationObject _de;

	/** Instanz des UmsetzungsModuls fr die Konvertierung der Telegramme in Attributgruppen (und zurck) */
	private DeConversionModul _conversionModul;

	/** Objektreferenz auf das bergeordnete Eak-Objekt. */
	private final Eak _eak;

	/** Hashtabelle, in der die auf diese DE referenzierenden Fachobjekte gespeichert werden. */

	private Hashtable _objectRefsToDe = new Hashtable();

	/** Name des UmsetzungsModul fr die Konvertierung der Telegramme in Attributgruppen (und zurck) */
	private String _conversionModulName;

	/** Bezeichnung des Gertes. */
	private final String _deDescription;

	/** Typ der DE-Daten. */
	private final boolean _isCluster;

	/** De Status. Wird abhngig von der De-Gutmeldung gesetzt */
	private boolean _isOk = false;

	/** De-Kanal-Nummer des DE-Blcks. */
	private final int _deChannel;

	/** Zugeordnete EA-Kanalnummer des DE am zugehrigen EAK. */
	private final int _eaChannel;

	/** Funktionsgruppe des DE. */
	private int _fg;

	/** Name des DE aus der Konfiguration */
	private String _dePid;

	// ------------------------------------------------------------------------
	// KONSTRUKTOREN  (und vom Konstruktor verwende Methoden)
	// ------------------------------------------------------------------------

	/** Erzeugt eine Instanz eines De. */
	public De(ConfigurationObject de, ClientDavInterface daf, Eak eak) {
		_de = de;
		_daf = daf;
		_eak = eak;

		DataModel config = _daf.getDataModel();

		AttributeGroup dePropertiesAtg = config.getAttributeGroup("atg.de");
		Data deProperties = _de.getConfigurationData(dePropertiesAtg);
		_dePid = de.getPid();
		_deDescription = deProperties.getTextValue("Bezeichnung").getText();
		_deChannel = deProperties.getScaledValue("DEKanal").intValue();
		_eaChannel = deProperties.getScaledValue("EAKanal").intValue();
		_isCluster = deProperties.getScaledValue("Cluster").getState().getName().equals("Ja");
		_conversionModulName = deProperties.getTextValue("UmsetzungsModul").getText().trim();

		// Ist DeKanal == 255, dann ist De immer ok, da sie niemals Daten erhlt, dient nur zum Senden
		if (_deChannel == 255) {
			_isOk = true;
		}

		// Funktionsgruppe des De ermitteln

		_fg = 0;
		if (_de.isOfType("typ.deLve")) {
			_fg = 1;
		}
		else if (_de.isOfType("typ.deAxl")) {
			_fg = 2;
		}
		else if (_de.isOfType("typ.deUfd")) {
			_fg = 3;
		}
		else if (_de.isOfType("typ.deWzg")) {
			_fg = 4;
		}
		else if (_de.isOfType("typ.deWww")) {
			_fg = 5;
		}
		else if (_de.isOfType("typ.deVlt")) {
			_fg = 6;
		}
		else if (_de.isOfType("typ.deZfr")) {
			_fg = 9;
		}
		else if (_de.isOfType("typ.deSys")) {
			_fg = 254;
		}
		else {
			for (int i = 128; i < 254; i++) {
				String deType = "typ.de" + i;
				if (_de.getType().getPid().equals(deType)) {
					_fg = i;
					break;
				}
			}
		}

		// De muss sich selbst registrieren und alle Objekte, die auf sie referenzieren, damit
		// diese vom UmsetzungsModul angemeldet werden knnen.

		putObjectRefToDe(_de);

		SystemObject[] objectsRefToDe = TlsApplicationLayer.getObjectRefsToDe(_de);
		for (int i = 0; i < objectsRefToDe.length; i++) {
			putObjectRefToDe(objectsRefToDe[i]);
		}

		// Wenn UmsetzungsModul nicht speziell konfiguriert wurde, dann abhngig
		// von der FG ein passendes StandardModul vorbelegen.
		if (_conversionModulName.equals("")) {
			switch (_fg) {
				case 1:
					_conversionModulName = "de.bsvrz.kex.tls.osi7.conversion.Fg001Default";
					break;
				case 2:
					_conversionModulName = "de.bsvrz.kex.tls.osi7.conversion.Fg002Default";
					break;
				case 3:
					_conversionModulName = "de.bsvrz.kex.tls.osi7.conversion.Fg003Default";
					break;
				case 4:
					_conversionModulName = "de.bsvrz.kex.tls.osi7.conversion.Fg004Default";
					break;
				case 5:
					_conversionModulName = "de.bsvrz.kex.tls.osi7.conversion.Fg005Default";
					break;
				case 6:
					_conversionModulName = "de.bsvrz.kex.tls.osi7.conversion.Fg006Default";
					break;
				case 9:
					_conversionModulName = "de.bsvrz.kex.tls.osi7.conversion.Fg009Default";
					break;
				case 254:
					_conversionModulName = "de.bsvrz.kex.tls.osi7.conversion.Fg254Default";
					break;
				default:
					_conversionModulName = "de.bsvrz.kex.tls.osi7.conversion.FgAllDefault";
					break;
			}
		}
		installConversionModule();

		debug.finest(toString());
	}

	/** !!!!Erzeugt eine TestInstanz eines De-Nur fr spezielle JUnit Tests verwenden!!!! */
	@SuppressWarnings({"AssignmentToNull"})
	public De(String deDescription, int deChannel, int eaChannel, boolean isCluster, int fg, String conversionModulName) {
		_de = null;
		_daf = null;
		_eak = null;

		_deDescription = deDescription;
		_deChannel = deChannel;
		_eaChannel = eaChannel;
		_isCluster = isCluster;
		_fg = fg;
		_conversionModulName = conversionModulName;

		// Ist DeKanal == 255, dann ist De immer ok, da sie niemals Daten erhlt, dient nur zum Senden
		if (_deChannel == 255) {
			_isOk = true;
		}

		// Wenn UmsetzungsModul nicht speziell konfiguriert wurde, dann abhngig
		// von der FG ein passendes StandardModul vorbelegen.
		if (_conversionModulName.equals("")) {
			switch (_fg) {
				case 1:
					_conversionModulName = "de.bsvrz.kex.tls.osi7.conversion.Fg001Default";
					break;
				case 2:
					_conversionModulName = "de.bsvrz.kex.tls.osi7.conversion.Fg002Default";
					break;
				case 3:
					_conversionModulName = "de.bsvrz.kex.tls.osi7.conversion.Fg003Default";
					break;
				case 4:
					_conversionModulName = "de.bsvrz.kex.tls.osi7.conversion.Fg004Default";
					break;
				case 5:
					_conversionModulName = "de.bsvrz.kex.tls.osi7.conversion.Fg005Default";
					break;
				case 6:
					_conversionModulName = "de.bsvrz.kex.tls.osi7.conversion.Fg006Default";
					break;
				case 9:
					_conversionModulName = "de.bsvrz.kex.tls.osi7.conversion.Fg009Default";
					break;
				case 254:
					_conversionModulName = "de.bsvrz.kex.tls.osi7.conversion.Fg254Default";
					break;
				default:
					_conversionModulName = "de.bsvrz.kex.tls.osi7.conversion.FgAllDefault";
					break;
			}
		}

		installConversionModule();

		debug.finest(toString());
	}

	/**
	 * Fgt ein Fachobjekt eines bestimmten Typs der Liste der auf dieses DE referenzierenden Objekte hinzu. Die Listen
	 * werden je Objekttyp getrennt verwaltet. Fr einen bestimmten Objekttyp (z.B. typ.fahrStreifen) kann die Liste der
	 * registrierten Objekte ber die Methode {@link #getObjectRefsToDe(SystemObjectType)} abgerufen werden.
	 *
	 * @param object Das auf diese DE referenzierende Fachobjekt.
	 */
	public void putObjectRefToDe(SystemObject object) {
		// ObjektTyp ermitteln, prfen ob fr diesen Typ bereits Eintrage vorliegen (gegebenfalls neue Liste anlegen),
		// und Objekt speichern.

		SystemObjectType objectType = object.getType();

		// Objekttyp von object ermitteln und dabei auf den Basis-De-Typen abbilden (sofern es sich um ein
		// typ.deXXX oder davon abgeleiteten Typ handelt).

		if (object.isOfType("typ.deLve")) {
			objectType = (SystemObjectType) _daf.getDataModel().getObject("typ.deLve");
		}
		else if (object.isOfType("typ.deAxl")) {
			objectType = (SystemObjectType) _daf.getDataModel().getObject("typ.deAxl");
		}
		else if (object.isOfType("typ.deUfd")) {
			objectType = (SystemObjectType) _daf.getDataModel().getObject("typ.deUfd");
		}
		else if (object.isOfType("typ.deWzg")) {
			objectType = (SystemObjectType) _daf.getDataModel().getObject("typ.deWzg");
		}
		else if (object.isOfType("typ.deWww")) {
			objectType = (SystemObjectType) _daf.getDataModel().getObject("typ.deWww");
		}
		else if (object.isOfType("typ.deVlt")) {
			objectType = (SystemObjectType) _daf.getDataModel().getObject("typ.deVlt");
		}
        else if (object.isOfType("typ.deZfr")) {
            objectType = (SystemObjectType) _daf.getDataModel().getObject("typ.deZfr");
        }
        else if (object.isOfType("typ.deSys")) {
            objectType = (SystemObjectType) _daf.getDataModel().getObject("typ.deSys");
        }
		else {
			for (int i = 128; i < 254; i++) {
				String deType = "typ.de" + i;
				if (object.getType().getPid().equals(deType)) {
					objectType = (SystemObjectType) _daf.getDataModel().getObject(deType);
					break;
				}
			}
		}

		ArrayList objects = (ArrayList) _objectRefsToDe.get(objectType);

		if (objects == null) {
			// neuen Eintrag anlegen
			objects = new ArrayList();
		}
		objects.add(object);
		_objectRefsToDe.put(objectType, objects);

	}

	private void installConversionModule() {
		try {
			// Anpassung zur Kompatibilitt nach Umstellung auf neue Packagestruktur de.bsvrz.*
			// Liegt Umsetzungsklasse noch im alten Package kex.tls.osi7.conversion.*, dann
			// wird automatisch im Package de.bsvrz.kex.tls.osi7.conversion.* gesucht.
			debug.fine("KonvertierungsModulName Vorgabe: [" + _conversionModulName + "]");
			if (Pattern.matches("^kex\\.tls\\.osi7\\.conversion\\..*", _conversionModulName)) {
				_conversionModulName = ("de.bsvrz." + _conversionModulName).trim();
			}
			debug.fine("KonvertierungsModulName nach Anpassung auf neue Packagestruktur de.bsvrz.*: [" + _conversionModulName + "]");

			// Umsetungsklasse laden und instanzieren.
			debug.finest("Versuche Umsetzungsklasse zu laden: " + _conversionModulName);
			_conversionClass = Class.forName(_conversionModulName);
			_conversionModul = (DeConversionModul) _conversionClass.getConstructor(new Class[]{ClientDavInterface.class, Eak.class, De.class}).newInstance(new Object[]{_daf, _eak, this});
		}
		catch (ClassNotFoundException e) {
			debug.error("Konnte UmsetzungsModul [" + _conversionModulName + "] nicht laden!", e);
		}
		catch (NoSuchMethodException e) {
			debug.error("Konstruktoraufruf in Klasse[" + _conversionModulName + "] fehlgeschlagen!", e);
		}
		catch (IllegalAccessException e) {
			debug.error("Zugriff auf UmsetzungsModul [" + _conversionModulName + "] fehlgeschlagen!", e);
		}
		catch (InstantiationException e) {
			debug.error("UmsetzungsModul [" + _conversionModulName + "] konnte  nicht instanziert werden!", e);
		}
		catch (InvocationTargetException e) {
			debug.error("Konstruktoraufruf am UmsetzungsModul [" + _conversionModulName + "] fehlgeschlagen!", e);
		}
		// Falls das Umsetzungsmodul aus einem der in den obigen catch-Blcken abgefangenen Grnde nicht instanziiert werden konnte, wird ein Dummy-Modul
		// instanziiert, um NullPointerExceptions bei spteren Zugriffen zu vermeiden.
		if(_conversionModul == null) {
			debug.warning("Statt dem UmsetzungsModul [" + _conversionModulName + "] wird ein Dummy-Umsetzungsmodul fr DE " + _dePid + " installiert!");
			_conversionModul = new DeConversionModul(_daf, _eak, this) {
				@Override
				public void processDeBlock(final DeBlock deBlock) {
					debug.warning("Empfangener DE-Block fr DE " + _dePid + " wird von Dummy-Umsetzungsmodul ignoriert, SollUmsetzungsmodul: " + _conversionModulName + ", DE-Block: " + deBlock);
				}
			};
		}
	}

	/**
	 * Liefert die String-Darstellung eines De:
	 * <pre><code>
	 * De (Bezeichnung)   :
	 * De-Typ             :
	 * De-Kanal-Nummer    :
	 * Funktionsgruppe    :
	 * Knotennummer       :
	 * EA-Kanal           :
	 * Ist Cluster-Kanal? :
	 * De-Status          :
	 * Umsetzungsmodul    :
	 * ObjektReferenzAufDe:    :
	 * </pre></code>
	 * Zustzlich werden alle angeschlossenen De mit ihren Eigenschften ausgegen.
	 *
	 * @return String-Darstellung
	 */

	public String toString() {
		StringBuffer sb = new StringBuffer();
		final String NEWLINE = System.getProperty("line.separator");
		final String INDENT2 = "        ";
		sb.append(INDENT2).append("........................................................................").append(NEWLINE);
		sb.append(INDENT2).append("De (Bezeichnung)   : ").append(getDeDescription()).append(NEWLINE);
		sb.append(INDENT2).append("De-Typ             : ").append(getObjectType()).append(NEWLINE);
		sb.append(INDENT2).append("De-Kanal-Nummer    : ").append(getDeChannel()).append(NEWLINE);
		sb.append(INDENT2).append("Funktionsgruppe    : ").append(getFg()).append(NEWLINE);
		sb.append(INDENT2).append("Knotennummer       : ");
		sb.append(getEak().getSm().getOsi7AdressDecimal()).append(" (Dez)   ");
		sb.append(getEak().getSm().getOsi7AdressHex()).append(" (Hex)   ");
		sb.append(getEak().getSm().getOsi7AdressLocCode()).append(" (LocCode)").append(NEWLINE);
		sb.append(INDENT2).append("EA-Kanal           : ").append(getEAChannel()).append(NEWLINE);
		sb.append(INDENT2).append("Ist Cluster-Kanal? : ").append(_isCluster ? "Ja" : "Nein").append(NEWLINE);
		sb.append(INDENT2).append("De-Status          : ").append(_isOk ? "ok" : "Fehler").append(NEWLINE);
		sb.append(INDENT2).append("Umsetzungsmodul    : ").append(_conversionModulName).append(NEWLINE);
		sb.append(INDENT2).append("ObjektReferenzAufDe: ").append(objectRefsToDeToString()).append(NEWLINE);
		return (sb.toString());
	}

	/**
	 * Zustzlichen Konstruktor fr DE,  die nicht ber die TLS-Versorgung sondern ber die Tabelle der zustzlichen
	 * Kommunikationspartner versorgt worden sind.
	 */
	public De(AdditionalCommunicationPartner acp, ClientDavInterface daf, Eak eak, Map conversionModules) {
		_de = acp.getReferencedObject();
		_daf = daf;
		_eak = eak;

		DataModel config = _daf.getDataModel();

		_dePid   = _de.getPid();
		_deDescription = "Virtuelles DE fr TLS hnliche Kommunikation";
		_deChannel = acp.getDeChannel();
		_eaChannel = -1;// wird bei TLS hnlichen Kommunikationspartner immer auf TLS-unzulssigen Wert gesetzt.
		_isCluster = false;//zustzliche Kommunikationspartner sind niemals Cluster.
		_isOk = true;//zustzliche Kommunikationspartner sind immer ok.
		_fg = acp.getFg();

		// Ermittlung der Umsetzungsscriptes fr diese DE. Ist am DE kein UmsetzungsModul konfiguriert,
		// wird der Standardeintrag verwendet (muss existieren!).

		_conversionModulName = acp.getConversionModulName().trim();
		if (_conversionModulName.equals("")) {
			try {
				_conversionModulName = ((String) conversionModules.get(_de.getType())).trim();
			}
			catch (NullPointerException e) {
				debug.error("UmsetzungsModuel fr DE (bei TLS hnlicher Kommunikation) muss \n" + "entweder direkt bei diesem Kommunikationspartner [" + acp.getReferencedObject().getNameOrPidOrId() + "] oder \n" + "in [atg.gertTlshnlicheKommunikationsPartnerUmsetzungsModule] \n" + "einmal fr den Typ [" + acp.getReferencedObject().getType().getPid() + "] fr das Gert\n" + "(VIZ, VRZ, UZ) angegeben werden!", e);
			}
		}

		// De muss sich selbst registrieren und alle Objekte, die auf sie referenzieren, damit
		// diese vom UmsetzungsModul angemeldet werden knnen.

		putObjectRefToDe(_de);

		SystemObject[] objectsRefToDe = TlsApplicationLayer.getObjectRefsToDe(_de);
		for (int i = 0; i < objectsRefToDe.length; i++) {
			putObjectRefToDe(objectsRefToDe[i]);
		}

		installConversionModule();

		debug.finest(toString());
	}

	// ------------------------------------------------------------------------
	// GETTER / SETTER METHODEN
	// ------------------------------------------------------------------------

	/**
	 * Liefert das Umsetzungsmodul fr dieses DE
	 *
	 * @return Instanz des UmsetzungsModuls fr die Konvertierung der Telegramme in Attributgruppen (und zurck)
	 */
	public DeConversionModul getConversionModul() {
		return _conversionModul;
	}

	/**
	 * Liefert das De-Konfigurationsobjekt aus der Konfiguration.
	 *
	 * @return De-Konfigurationsobjekt.
	 */
	public ConfigurationObject getDe() {
		return _de;
	}

	/**
	 * Liefert die De-Kanal-Nummer Dezimaldarstellung.
	 *
	 * @return De-Kanal-Nummer des De
	 */
	public int getDeChannel() {
		return _deChannel;
	}

	/**
	 * Liefert die Bezeichungstext des De.
	 *
	 * @return Bezeichnungstext des De
	 */
	public String getDeDescription() {
		return _deDescription;
	}

	/**
	 * Zugeordnete EA-Kanalnummer des DE am zugehrigen EAK.
	 *
	 * @return Zugeordnete EA-Kanalnummer des DE am zugehrigen EAK.
	 */
	public int getEAChannel() {
		return _eaChannel;
	}

	/**
	 * Liefert das bergeordnete Eak-Objekt.
	 *
	 * @return bergeordnetes Eak-Objekt.
	 */
	public Eak getEak() {
		return _eak;
	}

	/**
	 * Liefert die Funktionsgruppe dieses De.
	 *
	 * @return Funktionsgruppe der De.
	 */
	public int getFg() {
		return _fg;
	}

	/**
	 * Liefert einen Hashkey auf Basis der Osi7Adresse, der De-Kanal-Nummer und der Funktionsgruppe als String.
	 *
	 * @return Hashkey auf Basis der Osi7Adresse, der De-Kanal-Nummer und der Funktionsgruppe ("Osi7Adresse-DeNummer-FG",
	 *         z.B. "51043-193-4").
	 */
	public String getKey() {
		return (String.valueOf(String.valueOf(this.getEak().getSm().getOsi7Adress() + "-" + this.getDeChannel() + "-" + this.getFg())));
	}

	/**
	 * Testet, ob die De ein Clusterkanal ist.
	 *
	 * @return De ist Cluster.
	 */
	public boolean isCluster() {
		return _isCluster;
	}

	/**
	 * Testet den Zustand der De.
	 *
	 * @return <code>true</code>, wenn fr die De eine De-Gutmeldung (Typ 1, Id 1) empfangen wurde, ansonsten
	 *         <code>false</code>.
	 */
	public boolean isOk() {
		return _isOk;
	}

	/** Setzt den Zustand einer De */
	public void setOk(boolean isOk) {
		_isOk = isOk;
	}

	// ------------------------------------------------------------------------
	// SONSTIGE METHODEN
	// ------------------------------------------------------------------------

	/**
	 * Liefert die auf dieses DE referenzierenden Fachobjekte eines bestimmten Objekttyps und aller davon abgeleiteten
	 * Objekttypen. Dies knnen z.B. FahrstreifenObjekte sein, die Werte von dieser DE als Quelle erhalten.
	 *
	 * @param objectType ObjektTyp der Fachobjekte, deren Referenzen auf diese DE ermittelt werden sollen.
	 *
	 * @return Fachobjekte des angegebenen Objekttyps und aller davon abgeleiteten Typen (rekursiv), die auf diese De als
	 *         Quelle referenzieren (z.B. Fahrstreifen, die Werte von dieser DE erhatlen). Sind keien Objekte vorhanden,
	 *         wird ein leeres Array zurckgegeben.
	 */
	public SystemObject[] getObjectRefsToDe(SystemObjectType objectType) {
		List objects = new ArrayList();
		List objectTypes = new ArrayList();

		objectTypes.add(objectType);

		List subTypes = new ArrayList();
		subTypes = getSubTypes(objectType);

		objectTypes.addAll(subTypes);

		for (int i = 0; i < objectTypes.size(); i++) {
			SystemObjectType ot = (SystemObjectType) objectTypes.get(i);

			List objectList = (List) _objectRefsToDe.get(ot);
			if (objectList != null) {
				objects.addAll(objectList);
			}
		}

		if (objects == null) {
			return new SystemObject[0];
		}
		return (SystemObject[]) objects.toArray(new SystemObject[objects.size()]);
	}

	private List getSubTypes(SystemObjectType sot) {
		List subTypes = new ArrayList();
		subTypes.addAll(sot.getSubTypes());
		for (int i = 0; i < subTypes.size(); i++) {
			subTypes.addAll(getSubTypes((SystemObjectType) subTypes.get(i)));
		}
		return subTypes;
	}

	/**
	 * Liefert Namen/Pid/Id des Objekttyps diese Gertes.
	 *
	 * @return Objekttyps diese Gertes
	 */
	String getObjectType() {
		String s = "Unbekannter Gerte-Typ";
		s = _de.getType().getNameOrPidOrId();
		return s;
	}

	private String objectRefsToDeToString() {
		StringBuffer sb = new StringBuffer();
		Enumeration keys = _objectRefsToDe.keys();
		Object key;
		sb.append("{").append(_de.getPid()).append(":");
		while (keys.hasMoreElements()) {
			key = keys.nextElement();
			sb.append(((SystemObject) key).getNameOrPidOrId()).append(" [");
			SystemObject[] systemObjects = (SystemObject[]) ((ArrayList) _objectRefsToDe.get(key)).toArray(new SystemObject[0]);
			for (int i = 0; i < systemObjects.length; i++) {
				SystemObject systemObject = systemObjects[i];
				sb.append(i).append(":").append(systemObject.getNameOrPidOrId()).append(", ");
			}
			sb.append("];");
		}
		sb.append("}");

		return (sb.toString());
	}

	/** Verarbeitet einen DeBlock. */
	void processDeBlock(DeBlock deBlock) {
		_conversionModul.processDeBlock(deBlock);
	}

	public String getDePid() {
		return _dePid;
	}
}
