/*
 * Copyright (c) 2015 by inovat, innovative systeme - verkehr - tunnel - technik, Dipl.-Ing. H. C. Kniss
 *
 * This file is part of de.bsvrz.kex.tls.osi7.Sm
 *
 * de.bsvrz.kex.tls.osi7.Sm is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  de.bsvrz.kex.tls.osi7.Sm is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with de.bsvrz.kex.tls.osi7.Sm.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Diese Datei ist Teil von de.bsvrz.kex.tls.osi7.Sm.
 *
 *  de.bsvrz.kex.tls.osi7.Sm ist Freie Software: Sie knnen es unter den Bedingungen
 *  der GNU General Public License, wie von der Free Software Foundation,
 *  Version 3 der Lizenz oder (nach Ihrer Wahl) jeder spteren
 *  verffentlichten Version, weiterverbreiten und/oder modifizieren.
 *
 *  de.bsvrz.kex.tls.osi7.Sm wird in der Hoffnung, dass es ntzlich sein wird, aber
 *  OHNE JEDE GEWHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 *  Gewhrleistung der MARKTFHIGKEIT oder EIGNUNG FR EINEN BESTIMMTEN ZWECK.
 *  Siehe die GNU General Public License fr weitere Details.
 *
 *  Sie sollten eine Kopie der GNU General Public License zusammen mit diesem
 *  Programm erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * inovat, Dipl.-Ing. H. C. Kniss
 * An der Krautwiese 37
 * D-53783 Eitorf
 * +49 (0)2243 8464 193
 * info@inovat.de
 * www.inovat.de
 */

package de.bsvrz.kex.tls.osi7;

import de.bsvrz.dav.daf.main.*;
import de.bsvrz.dav.daf.main.config.*;
import de.bsvrz.sys.funclib.debug.Debug;

import java.util.*;

/**
 * Reprsentiert ein Steuermodul gem TLS.
 *
 * @author inovat, innovative systeme - verkehr - tunnel - technik
 * @author Dipl.-Ing. Hans Christian Kni (HCK)
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class Sm extends Device implements ClientReceiverInterface {
	// ------------------------------------------------------------------------
	// VARIABLENDEKLARATIONEN
	// ------------------------------------------------------------------------

	// Sonstige ---------------------------------------------------------------
	/** DebugLogger fr Debug-Ausgaben */
	private static final Debug debug = Debug.getLogger();

	/**
	 * Hashtabelle alle Eak. Die Identifikation erfolgt ber einen Schlssel auf "FG-DeNummer", z.B. "4-193". Jedes Eak
	 * wird fr alle FG-De Kombinationen eingetragen. Dadurch lt sich bei der Verarbeitung der (Einzel)-Telegramme das
	 * fr die Verarbeitung zustndige Eak einfach aus der FG des Einzeltelegramms und einer DE-Nummer (ungleich 255!) der
	 * enthaltenen DE-Blcke ermitteln.
	 */
	Map<String, Eak> _eakHashTable = new Hashtable<String, Eak>();

	/** Liste der angeschlossenen Eak. */
	List<Eak> _eakList = new ArrayList<Eak>();

	// ------------------------------------------------------------------------
	// KONSTRUKTOREN  (und vom Konstruktor verwende Methoden)
	// ------------------------------------------------------------------------

	/** Erzeugt eine Instanz eines SM (SteuerModul) */
	public Sm(ClientDavInterface daf, ConfigurationObject sm) {
		super(daf, sm);

		//Liste der angeschlossenen Eak ermitteln
		List eakList = sm.getObjectSet("Eak").getElements();
		Iterator eakIterator = eakList.iterator();
		while (eakIterator.hasNext()) {
			Eak eak = new Eak((ConfigurationObject) (eakIterator.next()), daf, this);
			//Liste der Eak aufbauen
			_eakList.add(eak);
			//Eak jetzt mit jedem mglichen FG-DE Schlssel in die Tabelle mit den Eak eintragen.
			String[] eakKeys = eak.getKeys();
			for (int i = 0; i < eakKeys.length; ++i) {
				_eakHashTable.put(eakKeys[i], eak);
			}
		}

		// Triggerdatenart fr Versand gepufferter DE anmelden
		DataModel config = daf.getDataModel();
		AttributeGroup atgSendSingleTelegram = config.getAttributeGroup("atg.tlsEinzelTelegrammSenden");
		Aspect aspect = config.getAspect("asp.tlsVorgabe");
		DataDescription dataDescription = new DataDescription(atgSendSingleTelegram, aspect);
		daf.subscribeReceiver(this, sm, dataDescription, new ReceiveOptions(false), new ReceiverRole(true));// Als Senke anmelden !!
		debug.finest(toString());
	}

	/**
	 * Liefert die String-Darstellung eines Gertes: Gibt zurert die allgemeine Darstellung eines Gertes aus. Zustzlich
	 * werden alle angeschlossenen Eak mit ihren Eigenschften ausgegen.
	 * <pre><code>
	 * Gert (Bezeichnung)             :
	 * Gertetyp                       :
	 * Knotennummer (Dez:Hex:Loc-Code) :
	 * Umsetzungsmodul                 :
	 * Schlsselwert (HashKey)         :
	 * Verbindungsstatus               :
	 * </pre></code>
	 *
	 * @return String-Darstellung
	 */
	public String toString() {
		StringBuffer sb = new StringBuffer();
		final String NEWLINE = System.getProperty("line.separator");
		sb.append(super.toString());
		sb.append("................................................................................").append(NEWLINE);
		Iterator i = _eakList.iterator();
		while (i.hasNext()) {
			sb.append(((i.next())).toString());
		}
		return (sb.toString());
	}

	/**
	 * Zustzlichen Konstruktor fr SM,  die nicht ber die TLS-Versorgung sondern ber die Tabelle der zustzlichen
	 * Kommunikationspartner versorgt worden sind.
	 */
	public Sm(ClientDavInterface daf, String osi7Adress, Map conversionModules) {
		super(daf, osi7Adress, conversionModules);
	}

	// ------------------------------------------------------------------------
	// INTERFACE METHODEN
	// ------------------------------------------------------------------------

	// --------------------- Interface ClientReceiverInterface ---------------------

	/**
	 * Aktualisierungsmethode, die nach Empfang eines angemeldeten Datensatzes von den
	 * Datenverteiler-Applikationsfunktionen aufgerufen wird. Diese Methode muss von der Applikation zur Verarbeitung der
	 * empfangenen Datenstze implementiert werden. Angemeldet ist hier nur die Attributgruppe
	 * atg.tlsEinzelTelegrammSenden
	 *
	 * @param results Feld mit den empfangenen Ergebnisdatenstzen.
	 */
	public void update(ResultData results[]) {
		// Abhngig vom Attributwert gespeicherte De-Blcke zum Versenden
		// bergeben oder Puffer lschen.
		for (ResultData result : results) {
			if (result.getData() != null) {
				if (result.getData().getUnscaledValue("Senden").getState().getName().equals("senden")) {
					// Fr alle EAK entsprechende Aktion auslsen
					for (Eak eak : _eakList) {
						eak.sendBufferedDeBlocks();
					}
				}
				else if (result.getData().getUnscaledValue("Senden").getState().getName().equals("verwerfen")) {
					// Fr alle EAK entsprechende Aktion auslsen
					for (Eak eak : _eakList) {
						eak.clearBufferedDeBlocks();
					}
				}
				else {
					debug.error("Unbekannter Wertezustand bei atg.tlsEinzelTelegrammSenden, Attribut 'Senden' erhalten (senden|verwerfen): " + result.getData().getUnscaledValue("Senden").getState().toString() + "\n an SM:\n" + this.toString());
				}
			}
		}
	}

	// ------------------------------------------------------------------------
	// SONSTIGE METHODEN
	// ------------------------------------------------------------------------

	public void addAdditionalCommunicationPartner(ClientDavInterface daf, List additionalDeviceTable, Map conversionModules) {
		Eak eak = new Eak(daf, this, additionalDeviceTable, conversionModules);
		//Liste der Eak aufbauen
		_eakList.add(eak);
		//Eak jetzt mit jedem mglichen FG-DE Schlssel in die Tabelle mit den Eak eintragen.
		String[] eakKeys = eak.getKeys();
		for (int i = 0; i < eakKeys.length; ++i) {
			_eakHashTable.put(eakKeys[i], eak);
		}
		debug.finest(toString());
	}

	/**
	 * Liefert das Eak an diesem SM mit der angegebenen Osi2Adresse.
	 *
	 * @param osi2Adress Osi2Adresse des Eak am Sm.
	 *
	 * @return Eak mit der gewnschten Osi2Adresse oder null, wenn Eak nicht vorhanden.
	 */
	public Eak getEak(int osi2Adress) {
		for (int i = 0; i < _eakList.size(); i++) {
			Eak eak = _eakList.get(i);
			if (eak.getOsi2Adress() == osi2Adress) {
				return eak;
			}
		}
		return null;
	}

	/**
	 * Verarbeitet ein Osi7Telegramm.
	 *
	 * @param osi7Telegram Das zu verarbeitende {@link Osi7Telegram}.
	 */
	void processOsi7Telegram(Osi7Telegram osi7Telegram) {
		// Zuerst spezielle Verarbeitung entsprechend dem geladenen Umsetzungsmodul durchfhren...
		_conversionModul.processOsi7Telegram(osi7Telegram);

		// ...danach EinzelTelegramm an die nchste Verarbeitungsstufe (Eak) weitergeben.
		for (int i = 0; i < osi7Telegram.getCountSingleTelegram(); ++i) {
			//Telegramm zur Verarbeitung an verantwortlichen Knoten delegieren
			SingleTelegram singleTelegram = osi7Telegram.getSingleTelegram()[i];

			Eak eak = null;
			for (int de : singleTelegram.getSetKeyDe()) {
				eak = _eakHashTable.get(singleTelegram.getKey(de));
				if (eak != null) break;
			}
			// Zustndiges EAK suchen...
			if (eak == null) {
				debug.warning(String.format("EinzelTelegramm kann nicht verarbeitet werden, da kein EAK zustndig: %s%n%s", singleTelegram.getDescriptionKey(), singleTelegram.toString()));
			}
			else {
				eak.processSingleTelegram(singleTelegram);
			}
		}
	}
}

