/*
 * Copyright (c) 2015 by inovat, innovative systeme - verkehr - tunnel - technik, Dipl.-Ing. H. C. Kniss
 *
 * This file is part of de.bsvrz.kex.tls.osi7.conversion.mv.Fg003Typ189Mv
 *
 * de.bsvrz.kex.tls.osi7.conversion.mv.Fg003Typ189Mv is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  de.bsvrz.kex.tls.osi7.conversion.mv.Fg003Typ189Mv is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with de.bsvrz.kex.tls.osi7.conversion.mv.Fg003Typ189Mv.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Diese Datei ist Teil von de.bsvrz.kex.tls.osi7.conversion.mv.Fg003Typ189Mv.
 *
 *  de.bsvrz.kex.tls.osi7.conversion.mv.Fg003Typ189Mv ist Freie Software: Sie knnen es unter den Bedingungen
 *  der GNU General Public License, wie von der Free Software Foundation,
 *  Version 3 der Lizenz oder (nach Ihrer Wahl) jeder spteren
 *  verffentlichten Version, weiterverbreiten und/oder modifizieren.
 *
 *  de.bsvrz.kex.tls.osi7.conversion.mv.Fg003Typ189Mv wird in der Hoffnung, dass es ntzlich sein wird, aber
 *  OHNE JEDE GEWHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 *  Gewhrleistung der MARKTFHIGKEIT oder EIGNUNG FR EINEN BESTIMMTEN ZWECK.
 *  Siehe die GNU General Public License fr weitere Details.
 *
 *  Sie sollten eine Kopie der GNU General Public License zusammen mit diesem
 *  Programm erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * inovat, Dipl.-Ing. H. C. Kniss
 * An der Krautwiese 37
 * D-53783 Eitorf
 * +49 (0)2243 8464 193
 * info@inovat.de
 * www.inovat.de
 */

package de.bsvrz.kex.tls.osi7.conversion.mv;

import de.bsvrz.dav.daf.main.*;
import de.bsvrz.kex.tls.osi7.*;
import de.bsvrz.kex.tls.osi7.conversion.Fg003Default;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Umsetzungsklasse fr FG 3, Typ 189, Firma Vaisala in Mecklenburg Vorpommern
 *
 * @author inovat, innovative systeme - verkehr - tunnel - technik / Logos GmbH
 * @author Dipl.-Ing. Hans Christian Kni (HCK) / Kristian Strey (KS)
 * @version 1.2 2004-12-16 (KS)
 */

public class Fg003Typ189Mv
		extends Fg003Default {
// ------------------------------------------------------------------------
// VARIABLENDEKLARATIONEN
// ------------------------------------------------------------------------

// Sonstige ---------------------------------------------------------------
	/**
	 * DebugLogger fr Debug-Ausgaben
	 */
	private static final Debug debug = Debug.getLogger();
// Private ----------------------------------------------------------------
	/**
	 * Tabelle, in der das Verhalten beim Empfang von Attributgruppen eingetragen wird. Jede Tabellenzeile hat folgenden
	 * Aufbau:
	 * <pre>
	 * "Attributgruppe (Pid)", "Aspekt (Pid)", "Fg", "Typ", "Id", "ObjektTyp"
	 * z.B.
	 * "atg.tlsGloKanalSteuerung", "asp.parameterSoll", "1", "29", "2", "typ.deLve"
	 * </pre>
	 * Die Auswertung erfolgt dabei wie folgt: <ul> <li> Fr jede Attributguppen/Aspekt Kombination erfolgt eine
	 * Empfangsanmeldung als Empfnger.</li> <li> In der Methode {@link #update} wird fr jede zur
	 * Attributgruppen/Aspket-Kombination passende Kombination eine Methode aufgerufen, deren Namen aus dem
	 * Tabelleneintrag wie folgt gebildet wird (hier fr den Beispieleintrag):<br> <code>Fg001Typ029Id002AtgTlsGloKanalSteuerungAspParameterSoll(ResultData)</code><br>
	 * Dabei enthlt <code>ResultData</code> die Daten der zu verarbeitenden ATG/ASP-Kombination.</li> <li> Fr jede
	 * Attributguppen/Aspekt Kombination erfolgt eine Empfangsanmeldung als Empfnger. Sind mehrere passende Eintrage
	 * vorhanden, werden entsprechend mehrere Methodenaufrufe durchgefhrt. Dadurch knnen mehrere ATG/ASP-Kombinationen
	 * in einen De-Block abgebildet werden (z.B. protokollnahe TLS-Attributgruppe und Attributgruppe eines
	 * Fachmodells):</li> <li> Ein Eintrag von "255" entspricht dabei einer "Wildcard" (passt immer). </ul>
	 */
	private final Object[][] davToTls = new Object[][]{
			// Attributgruppe, Aspekt, Fg, Typ, Id, ObjektTyp

	};

	/**
	 * Tabelle, in der das Verhalten beim Empfang von TLS De-Blcken eingetragen wird. Jede Tabellenzeile hat folgenden
	 * Aufbau:
	 * <pre>
	 * "Attributgruppe (Pid)", "Aspekt (Pid)", "Fg", "Typ", "Id", "ObjektTyp"
	 * z.B.
	 * "atg.tlsGloKanalSteuerung", "asp.tlsAntwort", "1", "29", "2", "typ.deLve"
	 * </pre>
	 * Die Auswertung erfolgt dabei wie folgt: <ul> <li> Fr jede Attributguppen/Aspekt Kombination erfolgt eine
	 * Sendeanmeldung als Quelle.</li> <li> In der Methode {@link #processDeBlock} wird fr jede zum De-Block passende
	 * Fg/Id/Typ-Kombination eine Methode aufgerufen, deren Namen aus dem Tabelleneintrag wie folgt gebildet wird (hier
	 * fr den Beispieleintrag):<br> <code>Fg001Typ029Id002AtgTlsGloKanalSteuerungAspTlsAntwort(DeBlock, Data,
	 * DataDescription)</code><br> Dabei enthlt der <code>deBlock</code> die Daten des zu verarbeitenden De-Blocks und
	 * <code>data</code> ein Datenobjekt der passenden Attributgruppen/Aspekt-Kombination, welchem direkt die Inhalte des
	 * De-Blocks zugewiesen werden knnen.<code>DataDescription</code> enthlt die passende {@link
	 * de.bsvrz.dav.daf.main.DataDescription} fr die zu sendenden Daten.</li> <li> Fr jede Attributguppen/Aspekt
	 * Kombination erfolgt eine Sendeanmeldung als Quelle. Sind mehrere passende Eintrage vorhanden, werden entsprechend
	 * mehrere Methodenaufrufe durchgefhrt. Dadurch kann ein De-Block mehrfach ausgewert und an verschiedene
	 * Attributgruppen weitergerleitet werden (z.B. protokollnahe TLS-Attributgruppe und Attributgruppe eine
	 * Fachmodells):</li> <li> Ein Eintrag von "255" entspricht dabei einer "Wildcard" (passt immer). </ul>
	 */
	private final Object[][] tlsToDav = new Object[][]{
			// Attributgruppe, Aspekt, Fg, Typ, Id, ObjektTyp
			{"atg.tlsUfdErgebnisMeldungZustandDerFahrbahnOberFlcheFBZVaisala", "asp.tlsAntwort", "3", "189", "4", "typ.deUfd"},
	};

// ------------------------------------------------------------------------
// KONSTRUKTOREN  (und vom Konstruktor verwende Methoden)
// ------------------------------------------------------------------------

	/**
	 * Objekterzeugung mit Anmeldung der globalen und lokalen Umsetzungseintrge
	 */
	public Fg003Typ189Mv(ClientDavInterface daf, Eak eak, De de) {
		super(daf, eak, de);
		subscribeAllData(tlsToDav, davToTls);
	}

// ------------------------------------------------------------------------
// SONSTIGE METHODEN
// ------------------------------------------------------------------------

	/**
	 * FG 3 : Typ: 189 Id: 4 Name: Herstellerspezifischer Fahrbahnzustand Vaisala (Antwort) TLS-Datenbeschreibung:<br>
	 * <pre>
	 * Position         Bezeichnung                             Erluterung
	 * Byte 1           Lnge DE-Block                          Lnge des folgenden DE-Blocks [ 5 ]
	 * Byte 2           Daten-Endgerte-Kanal                   [ 1..254 ]
	 * Byte 3           Typ der DE-Daten                        [ 189 ]
	 * Byte 4           Herstellercode                          (s. Anh. 7, Kap. 2)
	 * Byte 5           Herstellerversion                       siehe unten
	 * Byte 5           ZustandFahrbahnoberflcheVaisala        siehe unten
	 * </pre>
	 * Inhalt Byte: Herstellerversion ("Version"):
	 * <pre>
	 *    0-255:   Hilfsbyte fr die Dekodierhilfe - sie wird einzigartig vom Hersteller vergeben
	 * </pre>
	 * Inhalt Byte: ZustandFahrbahnoberflcheVaisala ("ZustandFahrbahnoberflcheVaisala"):
	 * <pre>
	 *        0:   Trocken (Vaisala Code 1 "dry" - entspricht Kodierung in Typ 70)
	 *       65:   Schnee, Glttegefahr (Vaisala Code 6 "snowy" - entspricht Kodierung in Typ 70)
	 *       66:   Eis, Glttegefahr (Vaisala Code 7 "icy" - entspricht Kodierung in Typ 70)
	 *       67:   Reif, Glttegefahr  (Vaisala Code 5 "frosty" - entspricht Kodierung in Typ 70)
	 *      128:   Feucht, Fahrbahntemperatur liegt ber Gefriertemperatur (Vaisala Code 2 "moist")
	 *      129:   Feucht, Restsalz vorhanden, Fahrbahntemperatur liegt ber Gefriertemperatur (Vaisala Code 8 "moist and
	 * chemical")
	 *      130:   Nass, Fahrbahntemperatur liegt ber Gefriertemperatur (Vaisala Code 3 "wet")
	 *      131:   Nass, Restsalz vorhanden, Fahrbahntemperatur liegt ber Gefriertemperatur (Vaisala Code 4 "wet and
	 * chemical")
	 *      255:   Wert nicht ermittelbar (Vaisala Code 0 "short")
	 * </pre>
	 *
	 * @param deBlock         Der zu bearbeitende De-Block.
	 * @param data            Der passende Datensatz fr die Ziel Attributgruppe/Aspekt-Kombination.
	 * @param dataDescription Die Datenbeschreibung fr die Attributgruppe/Aspekt-Kombination.
	 *
	 * @return Ergebnisdatensatz mit den vom DeBlock in die entsprechende Attributgruppe konvertierten Daten.
	 */
	public ResultData Fg003Typ189Id004AtgTlsUfdErgebnisMeldungZustandDerFahrbahnOberFlcheFBZVaisalaAspTlsAntwort(
			DeBlock deBlock, Data data, DataDescription dataDescription) {
		ResultData resultData = null;

		// _Messwerte_ aus Telegramm -----------------------------------------------------------------------------------
		int _Hersteller_ = deBlock.getDeData().readUnsignedByteAsInt();
		data.getUnscaledValue("Hersteller").set(_Hersteller_);

		int _Version_ = deBlock.getDeData().readUnsignedByteAsInt();
		data.getUnscaledValue("Version").set(_Version_);

		int _FBZ_ = deBlock.getDeData().readUnsignedByteAsInt();
		data.getUnscaledValue("ZustandFahrbahnoberflcheVaisala").set(_FBZ_);

		//Ergebnisdatensatz erzeugen...
		resultData = new ResultData(_de.getDe(), dataDescription, _eakConversionModul.getTimeStamp(), data);
		return resultData;
	}
}
